--------------------------------------------------------------------------------
--     This file is owned and controlled by Xilinx and must be used           --
--     solely for design, simulation, implementation and creation of          --
--     design files limited to Xilinx devices or technologies. Use            --
--     with non-Xilinx devices or technologies is expressly prohibited        --
--     and immediately terminates your license.                               --
--                                                                            --
--     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"          --
--     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR                --
--     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION        --
--     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION            --
--     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS              --
--     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,                --
--     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE       --
--     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY               --
--     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE                --
--     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR         --
--     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF        --
--     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS        --
--     FOR A PARTICULAR PURPOSE.                                              --
--                                                                            --
--     Xilinx products are not intended for use in life support               --
--     appliances, devices, or systems. Use in such applications are          --
--     expressly prohibited.                                                  --
--                                                                            --
--     (c) Copyright 1995-2007 Xilinx, Inc.                                     --
--     All rights reserved.                                                   --
--------------------------------------------------------------------------------
-- Company: Xilinx
-- Engineer: Michael Francis, michael.francis@xilinx.com
--
-- Module Name:   rs_dec_top.vhd
-- Project Name:  build
-- Target Device:  Virtex-4
-- Tool versions:  ISE 9.2.02
-- Description:   
-- The system consists of a the RS DEcoder. The decoder is clocked at otuclk, and once the data has been decoder it is
-- written to fifo. On the read side the data is read at oduclk and sent to the testbench and also writtem to a delay ram. 
-- There are two R-S decoders and each decoder has been set up for 8 channels. 


-----------------------------------
--Library declarations
-----------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;

-- for the primitives called up
LIBRARY unisim;
USE unisim.vcomponents.all;

-----------------------------------
-- Entity
-----------------------------------
ENTITY rs_dec_top IS
	port (
	 oduclk: IN std_logic;
	 otuclk: IN std_logic;
	 odu_reset: IN std_logic;
	 otu_reset: IN std_logic;
    data_in: IN std_logic_VECTOR(15 downto 0);
    enc_ready: IN std_logic;
	 sync_start: IN std_logic;
    data_out: OUT std_logic_VECTOR(15 downto 0);
	 err_found: OUT std_logic_vector(1 downto 0);
    err_cnt: OUT std_logic_VECTOR(9 downto 0);
    fail: OUT std_logic_vector(1 downto 0);
	 ready: OUT std_logic_vector(1 downto 0);
	 rffd: OUT std_logic_vector(1 downto 0);
	 rx_rd_en_valid: OUT std_logic);
END rs_dec_top ;

ARCHITECTURE rs_dec_a OF rs_dec_top IS


-----------------------------------
-- Component declarations
-----------------------------------
-----------------------------------
-- Reed Solomon Decoder
-----------------------------------
component rs_decoder
    port (
    data_in: IN std_logic_VECTOR(7 downto 0);
    sync: IN std_logic;
    reset: IN std_logic;
    clk: IN std_logic;
    ce: IN std_logic;
    data_out: OUT std_logic_VECTOR(7 downto 0);
    blk_strt: OUT std_logic;
    blk_end: OUT std_logic;
	 info_end: OUT std_logic;
    err_found: OUT std_logic;
    err_cnt: OUT std_logic_VECTOR(4 downto 0);
    fail: OUT std_logic;
    ready: OUT std_logic;
    rffd: OUT std_logic);
end component;


-----------------------------------
-- Async FIFO
-----------------------------------
component fifo
    port (
    din: IN std_logic_VECTOR(15 downto 0);
    rd_clk: IN std_logic;
    rd_en: IN std_logic;
    rst: IN std_logic;
    wr_clk: IN std_logic;
    wr_en: IN std_logic;
    dout: OUT std_logic_VECTOR(15 downto 0);
    empty: OUT std_logic;
    full: OUT std_logic;
    rd_data_count: OUT std_logic_VECTOR(1 downto 0);
    wr_data_count: OUT std_logic_VECTOR(1 downto 0));
end component;


CONSTANT PIPE:integer:=3+7;
-----------------------------------
-- Signals
-----------------------------------	
	SIGNAL vcc :std_logic:= '1';
	SIGNAL gnd :std_logic:='0';

	SIGNAL rx_full :std_logic;
	SIGNAL rx_empty :std_logic;
	SIGNAL rx_wr_en :std_logic;
	SIGNAL rx_rd_en :std_logic;
	SIGNAL rx_rd_en_d :std_logic;

	SIGNAL rs_data_out_reg :std_logic_vector(15 downto 0);
	SIGNAL rs_data_out :std_logic_vector(15 downto 0);

	SIGNAL rs_dec_blk_strt: std_logic_VECTOR(1 downto 0);
	SIGNAL rsdec_rffd: std_logic_VECTOR(1 downto 0):="00";
	SIGNAL sync :std_logic;	
	SIGNAL blk_strt :std_logic;	
	SIGNAL not_reset :std_logic;	
	SIGNAL sync_reset :std_logic;

	SIGNAL dec_info_end :std_logic_vector(1 downto 0);
	SIGNAL info_end_reg :std_logic;
	SIGNAL info_end :std_logic;
	
	

type regbank is array(PIPE downto 0) of std_logic;

SIGNAL reg: regbank;


BEGIN


---------------------------------------------------------------
-- Delay the start pulse so it can be used for the RS Decoder
---------------------------------------------------------------
reg(0) <= sync_start;

reg_bank: for i in 1 to PIPE generate begin
process (otuclk) begin
if otuclk'event and otuclk = '1' then	
reg(i) <= reg(i-1);
end if;
end process;
end generate;

sync <= reg(PIPE);

-----------------------------------------------------------------------
-- Reed Solomon Decoder. There are two dealing with 16 bit output from the
-- channel model Each Decoder is working on a byte.
-- The Decoder has been set to 8 channels
-----------------------------------------------------------------------

rs_dec: for i in 1 to 2 generate
      begin
dec : rs_decoder
        port map (
            data_in => data_in(i * 7 + (i - 1) downto ((i - 1) * 8)),
				sync => sync,
            reset => otu_reset,
            clk => otuclk,
            ce => vcc,
            data_out => rs_data_out(i * 7 + (i - 1) downto ((i - 1) * 8)),
            blk_strt => rs_dec_blk_strt(i-1),
            blk_end => open,
				info_end => dec_info_end(i-1),
            err_found => err_found(i-1),
            err_cnt => err_cnt(i * 4 + (i - 1) downto ((i - 1) * 5)),
            fail => fail(i-1),
            ready => ready(i-1),
            rffd => rsdec_rffd(i-1));
   end generate;		


-- AND the outputs from the two decoders				
	rffd <= 	rsdec_rffd;
	blk_strt <= rs_dec_blk_strt(0) and rs_dec_blk_strt(1);		
	info_end <= dec_info_end(0) and dec_info_end(1);
	
	
process(otuclk,otu_reset)
begin
if otu_reset ='1' then
	rs_data_out_reg <= (others =>'0');
elsif otuclk'event and otuclk = '1' then				
	rs_data_out_reg <= rs_data_out;
end if;
end process;					
				
				
--register the INFO_END signal so it ties up with the last data symbol					
process(otuclk,otu_reset)
begin
if otu_reset ='1' then
	info_end_reg <= '0';
elsif otuclk'event and otuclk = '1' then				
	info_end_reg <= info_end;
end if;
end process;				
				
	
--fifo write control generation
process(otuclk,otu_reset)
begin
if otu_reset ='1' then
	rx_wr_en <= '0';
elsif otuclk'event and otuclk = '1' then	
	if rx_full = '0' and blk_strt ='1' then
		rx_wr_en <= '1';
	elsif info_end_reg ='1' and info_end ='0' then
		rx_wr_en <= '0';
	end if;
	
	
end if;	
end process;
	
-----------------------------------
-- Receive Async FIFO
-----------------------------------				
	rx_fifo : fifo
        port map (
            din => rs_data_out_reg,
            wr_en => rx_wr_en,
            wr_clk => otuclk,
            rd_en => rx_rd_en,
            rd_clk => oduclk,
            rst => odu_reset,
            dout => data_out,
            full => rx_full,
            empty => rx_empty,
				wr_data_count => open,
            rd_data_count => open);
----------------------------------
-- Rx fifo controls				
-----------------------------------
--fifo read control
rx_rd_en<= not(rx_empty);
rx_rd_en_valid<= rx_rd_en;

	
END rs_dec_a;

