--------------------------------------------------------------------------------
--     This file is owned and controlled by Xilinx and must be used           --
--     solely for design, simulation, implementation and creation of          --
--     design files limited to Xilinx devices or technologies. Use            --
--     with non-Xilinx devices or technologies is expressly prohibited        --
--     and immediately terminates your license.                               --
--                                                                            --
--     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"          --
--     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR                --
--     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION        --
--     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION            --
--     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS              --
--     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,                --
--     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE       --
--     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY               --
--     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE                --
--     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR         --
--     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF        --
--     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS        --
--     FOR A PARTICULAR PURPOSE.                                              --
--                                                                            --
--     Xilinx products are not intended for use in life support               --
--     appliances, devices, or systems. Use in such applications are          --
--     expressly prohibited.                                                  --
--                                                                            --
--     (c) Copyright 1995-2007 Xilinx, Inc.                                   --
--     All rights reserved.                                                   --
--------------------------------------------------------------------------------
-- Company: Xilinx
-- Engineer: Michael Francis, michael.francis@xilinx.com
--
-- Design Name:   rs_top
-- Module Name:   rs_top_tb.vhd
-- Project Name:  build
-- Target Device:  Virtex-4
-- Tool versions:  ISE 9.2.02
-- Description:   
--
-- Notes: 
-- This ISE project is implementation files for XAPP Note 952(' Forward Error Correction on ITU G.709 
-- Networks using Reed-Solomon Solutions'. This test bench is for the testing of a Reed-Solomon(RS) encoder and decoder, and is 
-- targeted at the Virtex-4 family.  
-- The RS encoder and decoder have been configured for G.709 operation. This is a (255,239) configuration and means there
-- is the ability to correct up to (255-239)/2 symbol errors, where a symbol is a byte. That is a total of 8 bytes. 
--
-- In an OTU1 system the the OTU frame goes through byte interleaving to produce 16 FEC sub rows. 
-- This equates to data at 128 bits at 19.5 MHz. The sixteen sub rows are multiplexed into 16 bit data at 156 MHz. 
-- To represent this 128 bit random data is generated , multiplexed and written into a fifo at 156 MHz. In the design the 156 MHz
-- clock is called the ODUCLK. The ODU is the optical Data unit data. 
-- The fifo output data is read at OTUCLK at 167 MHz and sent to the RS encoder. Effectively the output rate of the fifo is 255/239 = 167 MHz. 
-- 239 bytes are read and the encoder appends 16 check symbols, or bytes.
-- The output from the encoder is sent to the decoder. The encoded data can , depending upon an error select signal, 
-- a) be passed through unchanged
-- b) be inverted and be within the error correcting capacity of the decoder,
-- c) be inverted and be outside the error correcting capacity of the decoder,
-- This is done via a case statement 
-- There is a constant 'RS_ERROR_START' that defines the start of errors and 'RS_ERROR_END' . By setting this to 8 or less then eventhough
-- there are errored symbols this is within the error correcting capability of the RS. If the number of errors is set to
-- 9 then the ' error' signal in the top level testbench goes to a '1'.
-- The RS decoder performs error correction , if the number of errors are within 8 symbols. 
-- The corrected data from the RS decoder is written to a FIFO and the output from the FIFO is the RS data output from the 
-- testbench
--
-- The testbench provides two output files;one for the transmitted data and one for the recieved data.

-----------------------------------
--Library declarations
-----------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.all;
USE std.textio.all;


-----------------------
-- Entity
-----------------------
ENTITY rs_top_tb_vhd IS
END rs_top_tb_vhd;

ARCHITECTURE behavior OF rs_top_tb_vhd IS 

-----------------------------------------------------------
	-- Component Declaration for the Unit Under Test (UUT)
-----------------------------------------------------------
	COMPONENT rs_top
	PORT(
		clk156_in: IN std_logic;
	   clk167_in: IN std_logic;
		reset_in : IN std_logic;      
		error_select: IN std_logic_vector(1 downto 0);-- define the type of error
		error : OUT std_logic);
	END COMPONENT;

	---------------------------------
	-- constant and signal declarations
   ---------------------------------
	--Inputs
	SIGNAL clk156 :  std_logic := '0';
	SIGNAL clk167 :  std_logic := '0';
	SIGNAL error_select :std_logic_vector(1 downto 0);--set this to 1 to enable errors

	
	SIGNAL reset :  std_logic := '1';


	--Outputs

	SIGNAL error :  std_logic;
	
	
BEGIN
--------------------------------------------
	-- Instantiate the Unit Under Test (UUT)
--------------------------------------------
	uut: rs_top PORT MAP(
		clk156_in => clk156,
		clk167_in => clk167,
		reset_in => reset,
		error_select => error_select,
		error => error);

---------------------------------
-- Clock signals
---------------------------------

clk156 <= not clk156 after 3.194 ns;
clk167 <= not clk167 after 2.994 ns;


---------------------------------
-- reset signals
---------------------------------

reset <=  '0' after 500 ns;



	
	tb:process
begin 
error_select <= "00"; -- no errors
wait for 50 us;
error_select <= "01"; -- inverted data and max 8 symbol errors per channel
wait for 50 us;
error_select <= "10"; -- inverted data and max 9 symbol errors per channel. output error signal goes active
wait for 50 us;
end process;
	
END;
