--------------------------------------------------------------------------------
--     This file is owned and controlled by Xilinx and must be used           --
--     solely for design, simulation, implementation and creation of          --
--     design files limited to Xilinx devices or technologies. Use            --
--     with non-Xilinx devices or technologies is expressly prohibited        --
--     and immediately terminates your license.                               --
--                                                                            --
--     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"          --
--     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR                --
--     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION        --
--     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION            --
--     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS              --
--     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,                --
--     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE       --
--     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY               --
--     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE                --
--     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR         --
--     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF        --
--     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS        --
--     FOR A PARTICULAR PURPOSE.                                              --
--                                                                            --
--     Xilinx products are not intended for use in life support               --
--     appliances, devices, or systems. Use in such applications are          --
--     expressly prohibited.                                                  --
--                                                                            --
--     (c) Copyright 1995-2007 Xilinx, Inc.                                   --
--     All rights reserved.                                                   --
--------------------------------------------------------------------------------
-- Company: Xilinx
-- Engineer: Michael Francis, michael.francis@xilinx.com
--
-- Design Name:   rs_top
-- Module Name:   rs_top_tb.vhd
-- Project Name:  build
-- Target Device:  Virtex-5
-- Tool versions:  ISE 9.2.02
-- Description:   
--
-- Notes: 
-- This ISE project is implementation files for XAPP Note 952(' Forward Error Correction on ITU G.709 
-- Networks using Reed-Solomon Solutions'. This test bench is for the testing of a Reed-Solomon(RS) encoder and decoder. 
-- The RS encoder and decoder have been configured for G.709 operation. This is a (255,239) configuration and means there
-- is the ability to correct up to (255-239)/2 symbol errors, where a symbol is a byte. That is a total of 8 bytes. 
--
-- In an OTU2 system the the OTU frame goes through byte interleaving to produce 16 FEC sub rows. 
-- This is represented in this project as random data at 128 bits at 78 MHz, and is multiplexed into 64 bit data at 156 MHz. 
-- which is written into a fifo. This represents two channels multipliexed into each encoder. 
-- The fifo output data  is read at 167 MHz and sent to the two channel RS encoder. Effectively the output rate of the fifo 
-- is 255/239 = 167 MHz. 239 bytes are read and the encoder appends 16 check symbols, or bytes.
-- The output from the encoder is sent to the decoder. The encoded data can , depending upon
-- an error select signal, 
-- a) be passed through unchanged
-- b) be inverted and be within the error correcting capacity of the decoder,
-- c) be inverted and be outside the error correcting capacity of the decoder,
-- The RS decoder performs error correction , if the number of errors are within 8 symbols. The corrected data from the
-- RS decoder is written to a FIFO and the output from the FIFO is the RS data output from the 
-- testbench. The testbench provides an error signal that compares the transmitted data and the recieved data.
-- If there are errors this is indicated by error signal


-----------------------------------
--Library declarations
-----------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.all;
USE std.textio.all;


-----------------------
-- Entity
-----------------------
ENTITY rs_top_tb_vhd IS
END rs_top_tb_vhd;

ARCHITECTURE behavior OF rs_top_tb_vhd IS 

-----------------------------------------------------------
	-- Component Declaration for the Unit Under Test (UUT)
-----------------------------------------------------------
	COMPONENT rs_top
	PORT(
	clk200_in_N: IN std_logic;
	clk200_in_P: IN std_logic;
	reset_n : IN std_logic;      
	error_select: IN std_logic_vector(1 downto 0);-- define the type of error
	error : OUT std_logic;-- the tx data and rx data are not the same. 
	GPIO_LED_5: OUT std_logic);
	END COMPONENT;

	---------------------------------
	-- constant and signal declarations
   ---------------------------------
	--Inputs
	SIGNAL clk200_in_N: std_logic := '0';
	SIGNAL clk200_in_P: std_logic := '1';
	SIGNAL reset :  std_logic := '1';
	SIGNAL error_select :std_logic_vector(1 downto 0);--set this to 1 to enable errors


	--Outputs
	
	SIGNAL info :  std_logic;
	SIGNAL error :  std_logic;
	SIGNAL GPIO_LED_5 :  std_logic;


	
	
BEGIN
--------------------------------------------
	-- Instantiate the Unit Under Test (UUT)
--------------------------------------------
	uut: rs_top PORT MAP(
		clk200_in_N => clk200_in_N ,
		clk200_in_P => clk200_in_P ,
		reset_n => reset,
		error_select => error_select,
		error => error,
		GPIO_LED_5 => GPIO_LED_5);

---------------------------------
-- Clock signals
---------------------------------

clk200_in_N <= not clk200_in_N after 2.5 ns;
clk200_in_P <= not clk200_in_P after 2.5 ns;

---------------------------------
-- reset signals
---------------------------------

reset <=  '1' after 200 ns;

tb:process
begin 
error_select <= "00"; -- no errors
wait for 50 us;
error_select <= "01"; -- inverted data and max 8 symbol errors per channel
wait for 50 us;
error_select <= "10"; -- inverted data and max 9 symbol errors per channel. output error signal goes active
wait for 50 us;
end process;

	
	
END;
