--------------------------------------------------------------------------------
--     This file is owned and controlled by Xilinx and must be used           --
--     solely for design, simulation, implementation and creation of          --
--     design files limited to Xilinx devices or technologies. Use            --
--     with non-Xilinx devices or technologies is expressly prohibited        --
--     and immediately terminates your license.                               --
--                                                                            --
--     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"          --
--     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR                --
--     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION        --
--     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION            --
--     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS              --
--     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,                --
--     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE       --
--     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY               --
--     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE                --
--     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR         --
--     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF        --
--     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS        --
--     FOR A PARTICULAR PURPOSE.                                              --
--                                                                            --
--     Xilinx products are not intended for use in life support               --
--     appliances, devices, or systems. Use in such applications are          --
--     expressly prohibited.                                                  --
--                                                                            --
--     (c) Copyright 1995-2007 Xilinx, Inc.                                   --
--     All rights reserved.                                                   --
--------------------------------------------------------------------------------
-- Company: Xilinx
-- Engineer: Michael Francis, michael.francis@xilinx.com
--
-- Design Name:   rs_top
-- Module Name:   rs_top_tb.vhd
-- Project Name:  build
-- Target Device:  Virtex-4
-- Tool versions:  ISE 9.2.02
-- Description:   
-- The system consists of a random source and a RS Encoder. The random source is clocked at 156 MHz, and is 
-- written to an async fifo. On the read side the data is read at 167 MHz and sent to the RS Encoder(255,239) IP core. 



-----------------------------------
--Library declarations
-----------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;

-- for the primitives called up
LIBRARY unisim;
USE unisim.vcomponents.all;



-----------------------------------
-- Entity
-----------------------------------
ENTITY rs_enc_top IS
	port (
	 oduclk: IN std_logic;
	 otuclk: IN std_logic;
	 odu_reset: IN std_logic;
	 otu_reset: IN std_logic;
    data_in: IN std_logic_VECTOR(255 downto 0);
    bypass: IN std_logic;
	 enc_ready : OUT std_logic;
	 sync_start: OUT std_logic;
	 tx_wr_en : OUT std_logic;
    data_out: OUT std_logic_VECTOR(255 downto 0);
    info: OUT std_logic);
END rs_enc_top ;

ARCHITECTURE rs_enc_a OF rs_enc_top IS

-----------------------------------
-- Reed Solomon Encoder
-----------------------------------
component rs_encoder
    port (
    data_in: IN std_logic_VECTOR(7 downto 0);
    start: IN std_logic;
    bypass: IN std_logic;
    nd: IN std_logic;
    data_out: OUT std_logic_VECTOR(7 downto 0);
    info: OUT std_logic;
    rdy: OUT std_logic;
    rfd: OUT std_logic;
    rffd: OUT std_logic;
    ce: IN std_logic;
    reset: IN std_logic;
    clk: IN std_logic);
end component;

-----------------------------------
-- Async FIFO
-----------------------------------

component fifo
    port (
    din: IN std_logic_VECTOR(255 downto 0);
    rd_clk: IN std_logic;
    rd_en: IN std_logic;
    rst: IN std_logic;
    wr_clk: IN std_logic;
    wr_en: IN std_logic;
    dout: OUT std_logic_VECTOR(255 downto 0);
    empty: OUT std_logic;
    full: OUT std_logic;
    rd_data_count: OUT std_logic_VECTOR(1 downto 0);
    wr_data_count: OUT std_logic_VECTOR(1 downto 0));
end component;

-----------------------------------
-- Signals
-----------------------------------	

	SIGNAL vcc :std_logic:= '1';
	SIGNAL gnd :std_logic:='0';
	SIGNAL tx_wr_en_fifo :std_logic;
	SIGNAL tx_rd_en :std_logic;
	SIGNAL tx_full :std_logic;
	SIGNAL tx_empty :std_logic;
	SIGNAL tx_go :std_logic:='0';
	SIGNAL enc_start :std_logic;
	SIGNAL wr_fifo_dout :std_logic_VECTOR(255 downto 0);

-------------------------------	
-- RS Encoder Control signals
-------------------------------	

	SIGNAL rs_enc_rdy :std_logic_vector(31 downto 0);
	SIGNAL rs_enc_rfd :std_logic_vector(31 downto 0);
	SIGNAL rs_enc_rffd :std_logic_vector(31 downto 0);
	SIGNAL not_reset :std_logic;	
	SIGNAL sync_reset :std_logic;
	SIGNAL enc_rfd :std_logic;	
	SIGNAL enc_rffd :std_logic;	
	SIGNAL tx_rd_count: std_logic_VECTOR(1 downto 0);

	SIGNAL rs_enc_info :std_logic_vector(31 downto 0);
	SIGNAL first_start :std_logic;
	SIGNAL start :std_logic;


BEGIN

-----------------------------------
-- Transmit Async FIFO
-----------------------------------
tx_fifo : fifo
        port map (
            din => data_in,
            wr_en => tx_wr_en_fifo,
            wr_clk => oduclk,
				rd_en => tx_rd_en,							
            rd_clk => otuclk,
            rst => odu_reset,
            dout => wr_fifo_dout,
            full => tx_full,
            empty => tx_empty,
				wr_data_count => open,
            rd_data_count => tx_rd_count);

				
----------------------------------
-- Tx fifo controls				
-----------------------------------
--fifo write control
tx_wr_en_fifo <= not (tx_full);
tx_wr_en <= not (tx_full);-- send this to the outside. 


--fifo read control
process(otuclk,otu_reset)
begin
if otu_reset ='1' then
	tx_go <= '0';
elsif otuclk'event and otuclk = '1' then	
	if tx_rd_count > "01"  then--lets wait until the fifo is half full before reading data
		tx_go <= '1';
	end if;
end if;		
end process;

-- encoder control

tx_rd_en <= enc_rfd and tx_go;

process(otuclk,otu_reset)
begin
if otu_reset ='1' then
 start <='0';
 first_start <= '0';
elsif otuclk'event and otuclk = '1' then

   if tx_rd_en = '1'  and enc_rffd = '1' then

			start <= '1';
	else
			start <= '0';
	end if;	
	
	--look for the when the first block is read
	if tx_rd_en = '1'  and enc_rffd = '1'then

			first_start <= '1';
	end if;	
	
end if;
end process;
--------------------------------------------------------------------------
-- Reed Solomon Encoder. There are two dealing with 16 bit output from the
-- FIFOs. Each Encoder is working on a byte. The Encoder has been set to 8 channels
--------------------------------------------------------------------------

rs: for i in 1 to 32 generate
      begin
enc : rs_encoder
        port map (
				data_in => wr_fifo_dout(i * 7 + (i - 1) downto ((i - 1) * 8)),
            start => enc_start,
            bypass => bypass,
            nd => vcc,
            data_out => data_out(i * 7 + (i - 1) downto ((i - 1) * 8)),
            info => rs_enc_info(i-1),
            rdy => rs_enc_rdy(i-1),
            rfd => rs_enc_rfd(i-1),
            rffd => rs_enc_rffd(i-1),
				ce => vcc,
            reset => otu_reset,
            clk => otuclk);
   end generate;	
			
			
info <= rs_enc_info(1) and rs_enc_info(0);
enc_ready <= rs_enc_rdy(0) and rs_enc_rdy(1);
enc_rfd <= rs_enc_rfd(0) and rs_enc_rfd(1);
enc_rffd <= rs_enc_rffd(0) and rs_enc_rffd(1);

enc_start <= (enc_rffd and start) or (enc_rffd and first_start);

sync_start <= enc_start; --passed onto to the decoder
				
END rs_enc_a;

