--------------------------------------------------------------------------------
--     This file is owned and controlled by Xilinx and must be used           --
--     solely for design, simulation, implementation and creation of          --
--     design files limited to Xilinx devices or technologies. Use            --
--     with non-Xilinx devices or technologies is expressly prohibited        --
--     and immediately terminates your license.                               --
--                                                                            --
--     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"          --
--     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR                --
--     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION        --
--     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION            --
--     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS              --
--     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,                --
--     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE       --
--     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY               --
--     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE                --
--     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR         --
--     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF        --
--     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS        --
--     FOR A PARTICULAR PURPOSE.                                              --
--                                                                            --
--     Xilinx products are not intended for use in life support               --
--     appliances, devices, or systems. Use in such applications are          --
--     expressly prohibited.                                                  --
--                                                                            --
--     (c) Copyright 1995-2007 Xilinx, Inc.                                   --
--     All rights reserved.                                                   --
--------------------------------------------------------------------------------
-- Company: Xilinx
-- Engineer: Michael Francis, michael.francis@xilinx.com
--
-- Design Name:   rs_top
-- Module Name:   rs_top_tb.vhd
-- Project Name:  build
-- Target Device:  Virtex-4
-- Tool versions:  ISE 9.2.02
-- Description:   
-- This is an example for a OTU1 system. The system consists of a random source. 
-- This is clocked at 156 MHz. The data is written to an async fifo. On the read side the data is read
-- at 167 MHz and sent to the RS(255,239) IP core. The check symbols are appended and the data is sent through 
-- error channel to the RS decoder where the k data symbols are written to the fifo. This is read out at 156 MHz.
-- There is a constant 'RS_ERROR_NUMBER' that defines the number of errors. By setting this to 8 or less then eventhough
-- there are errored symbols this is within the error correcting capability of the RS. If the number of errors is set to
-- 9 then the ' error' signal in the top level testbench goes to a '1'. 
-- The testbench provides two output files;one for the transmitted data and one for the recieved data.



-----------------------------------
--Library declarations
-----------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;

-- for the primitives called up
LIBRARY unisim;
USE unisim.vcomponents.all;

-----------------------------------
-- Entity
-----------------------------------
ENTITY rs_top IS
	port (
--inputs
		clk156_in: IN std_logic;
	   clk167_in: IN std_logic;
		reset_in : IN std_logic;      
		error_select: IN std_logic_vector(1 downto 0);-- define the type of error
		error : OUT std_logic);
END rs_top ;

ARCHITECTURE rs_top_a OF rs_top IS


-----------------------------------
-- Random Data Generator
-----------------------------------
component random_gen
	port (
	 oduclk: IN std_logic;
	 reset: IN std_logic;
	 en : IN std_logic;
    randomdata_init: IN std_logic_VECTOR(35 downto 0);	 
	 randomdata: OUT std_logic_VECTOR(31 downto 0));
end component;


-----------------------------------
-- RS Encoder top
-----------------------------------
component rs_enc_top
	port (
	 oduclk: IN std_logic;
	 otuclk: IN std_logic;
	 odu_reset: IN std_logic;
	 otu_reset: IN std_logic;
    data_in: IN std_logic_VECTOR(255 downto 0);
    bypass: IN std_logic;
	 enc_ready : OUT std_logic;
	 sync_start: OUT std_logic;
	 tx_wr_en : OUT std_logic;
    data_out: OUT std_logic_VECTOR(255 downto 0);
    info: OUT std_logic);
end component;



-----------------------------------
-- RS Decoder top
-----------------------------------
component rs_dec_top
	port (
	 oduclk: IN std_logic;
	 otuclk: IN std_logic;
	 odu_reset: IN std_logic;
	 otu_reset: IN std_logic;
    data_in: IN std_logic_VECTOR(255 downto 0);
    enc_ready: IN std_logic;
	 sync_start: IN std_logic;
    data_out: OUT std_logic_VECTOR(255 downto 0);
	 err_found: OUT std_logic_vector(31 downto 0);
    err_cnt: OUT std_logic_VECTOR(159 downto 0);
    fail: OUT std_logic_vector(31 downto 0);
	 ready: OUT std_logic_vector(31 downto 0);
	 rffd: OUT std_logic_vector(31 downto 0);
	 rx_rd_en_valid: OUT std_logic);
end component;


-----------------------------------
-- DCM
-----------------------------------



	COMPONENT dcm_oduclk
	PORT(
		CLKIN_IN : IN std_logic;
		RST_IN : IN std_logic;          
		CLKIN_IBUFG_OUT : OUT std_logic;
		CLK0_OUT : OUT std_logic;
		CLK2X_OUT : OUT std_logic;
		LOCKED_OUT : OUT std_logic
		);
	END COMPONENT;
	
	COMPONENT dcm_otuclk
	PORT(
		CLKIN_IN : IN std_logic;
		RST_IN : IN std_logic;          
		CLKIN_IBUFG_OUT : OUT std_logic;
		CLK0_OUT : OUT std_logic;
		CLK2X_OUT : OUT std_logic;
		LOCKED_OUT : OUT std_logic
		);
	END COMPONENT;
		
	
-----------------------------------
-- Signals
-----------------------------------	
	SIGNAL reset :std_logic:='1';
	SIGNAL oduclk :std_logic:='0';
	SIGNAL otuclk :std_logic:='0';
	SIGNAL oduclk2x :std_logic:='0';
	SIGNAL otuclk2x :std_logic:='0';
	SIGNAL lock_dcm0 :std_logic:='0';
	SIGNAL lock_dcm1 :std_logic:='0';
	SIGNAL vcc :std_logic:= '1';
	SIGNAL gnd :std_logic:='0';
	SIGNAL odu_rst_sr: std_logic_vector(31 downto 0):=(others=>'1');
	SIGNAL otu_rst_sr: std_logic_vector(31 downto 0):=(others=>'1');
   SIGNAL lock_dcm1_not :std_logic;
	SIGNAL lock_dcm0_not :std_logic;
	SIGNAL odu_reset :std_logic:='1';
	SIGNAL otu_reset :std_logic:='1';
	
-----------------------------------
-- RS Decoder Signals
-----------------------------------		
	 SIGNAL rs_dec_data_in: std_logic_VECTOR(255 downto 0);
	 SIGNAL sync_start: std_logic;
    SIGNAL rs_data_out,rs_data_out_d: std_logic_VECTOR(255 downto 0);
	 SIGNAL err_found: std_logic_vector(31 downto 0);
    SIGNAL err_cnt: std_logic_VECTOR(159 downto 0);
    SIGNAL fail: std_logic_vector(31 downto 0);
	 SIGNAL ready: std_logic_vector(31 downto 0);
	 SIGNAL rffd: std_logic_vector(31 downto 0);
	 SIGNAL rx_rd_en_v: std_logic;
    SIGNAL rx_rd_en_v_d,rx_rd_en_v_dd :std_logic;	
	

-----------------------------------
-- RS Encoder Signals
-----------------------------------	

	SIGNAL bypass :std_logic:='0';	
	SIGNAL enc_ready : std_logic;
   SIGNAL rs_encode_data: std_logic_VECTOR(255 downto 0);
   SIGNAL info: std_logic;
	SIGNAL tx_wr_en :std_logic;	


   SIGNAL delay_source,delay_source_d :std_logic_VECTOR(255 downto 0);
	SIGNAL FEC_sub_row_data_lsb :std_logic_VECTOR(127 downto 0);	

	SIGNAL count : std_logic_vector(2 downto 0);
	SIGNAL en :std_logic:='0';	
   SIGNAL randomdata_init: std_logic_VECTOR(143 downto 0):=X"1157FF7E4EE57587E4AA57581145557007E4";
	SIGNAL randomdata: std_logic_VECTOR(127 downto 0):=(others =>'0');
	SIGNAL FEC_sub_row_data,FEC_sub_row_data_d : std_logic_VECTOR(255 downto 0):=(others =>'0');	
   TYPE fec_data_array is array (15 downto 0) of STD_LOGIC_VECTOR (7 downto 0);
	SIGNAL FEC_sub_row_array:fec_data_array;
	
	

-----------------------------------
-- Error control Signals
-----------------------------------	

	SIGNAL send_error :std_logic;
   SIGNAL error_en :std_logic;
	SIGNAL RS_ERROR_START : std_logic_vector(7 downto 0):= "00000000";
	SIGNAL RS_ERROR_END : std_logic_vector(7 downto 0):= "00000000";
	SIGNAL err_count : std_logic_vector(7 downto 0):= "00000000";	
	SIGNAL error_int :std_logic_vector(3 downto 0);
-----------------------------------
-- delay ram control Signals
-----------------------------------		
	SIGNAL write_addr,read_addr,read_addr_b: std_logic_vector(11 downto 0):=(others=>'0');

-- Define the ram array used to delay the transmited data 
	TYPE delay_ram_array is array (4095 downto 0) of std_logic_vector (255 downto 0);
	SIGNAL delay_ram: delay_ram_array;

	attribute max_fanout: string;
	attribute max_fanout of en: signal is "30";
	
BEGIN


-----------------------------------
-- DCMs for 156 MHz and 167 MHz
-----------------------------------
	
		Inst_dcm_oduclk: dcm_oduclk PORT MAP(
		CLKIN_IN => clk156_in,
		RST_IN => reset_in,
		CLKIN_IBUFG_OUT => open,
		CLK0_OUT => oduclk,
		CLK2X_OUT => oduclk2x,
		LOCKED_OUT => lock_dcm0	);
		
		
		Inst_dcm_otuclk: dcm_otuclk PORT MAP(
		CLKIN_IN => clk167_in,
		RST_IN => reset_in,
		CLKIN_IBUFG_OUT => open,
		CLK0_OUT => otuclk,
		CLK2X_OUT => otuclk2x,
		LOCKED_OUT => lock_dcm1	);


-----------------------------------
-- Delay to create reset
-----------------------------------		 
lock_dcm0_not<= not lock_dcm0;
lock_dcm1_not<= not lock_dcm1;
-----------------------------------
-- Delay to create odu_reset
-----------------------------------		
process (oduclk)
begin
   if oduclk'event and oduclk ='1' then  
      for i in 0 to 29 loop
         odu_rst_sr(i+1) <= odu_rst_sr(i);
      end loop;
      odu_rst_sr(0) <= lock_dcm0_not;
   end if;
end process;

odu_reset<= odu_rst_sr(30);

-----------------------------------
-- Delay to create otu_reset
-----------------------------------	
process (otuclk)
begin
   if otuclk'event and otuclk ='1' then  
      for i in 0 to 29 loop
         otu_rst_sr(i+1) <= otu_rst_sr(i);
      end loop;
      otu_rst_sr(0) <= lock_dcm1_not;
   end if;
end process;

otu_reset<= otu_rst_sr(30);


process(oduclk2x)
begin
if (oduclk2x = '1' and oduclk2x'event) then
      if (odu_reset = '1') then
         count <= (others =>'0');
		else
         count <= count + 1;
		end if;
		
		if (count(0) = '1') then
         en<= '1';
      else
         en<= '0';
		end if;
		
end if;					
end process;

----------------------------------------------------------------------
-- Random data. Create 4 instances to create 128 bits
-- Data is generated at 312 MHZ
----------------------------------------------------------------------

rd: for i in 1 to 4 generate
      begin
rand :random_gen
        port map (
            oduclk => oduclk2x,
            reset => odu_reset,
            en => VCC,
            randomdata_init => randomdata_init(i * 35 + (i - 1) downto ((i - 1) * 36)),
            randomdata => randomdata(i * 31 + (i - 1) downto ((i - 1) * 32)));    
   end generate;	  
 
 
----------------------------------------------------------------------
-- Demux the 128-bit random data @ 312 MHZ into 255 random data @ 156 MHZ
----------------------------------------------------------------------
process(oduclk2x)
begin
if (oduclk2x = '1' and oduclk2x'event) then
      if (en = '1') then
         FEC_sub_row_data_lsb <= randomdata;
		end if;
		if (en = '0') then
         FEC_sub_row_data <= FEC_sub_row_data_lsb & randomdata;
		end if;
		FEC_sub_row_data_d <= FEC_sub_row_data;
		end if;					
end process;
 

 
-----------------------------------
-- RS Encoder and FIFO
-----------------------------------			

rs_encoder_top: rs_enc_top
port map (
	 oduclk => oduclk,
	 otuclk => otuclk,
	 odu_reset => odu_reset,
	 otu_reset => otu_reset,
    data_in => FEC_sub_row_data,
    bypass => bypass,
	 sync_start => sync_start,
	 enc_ready =>enc_ready,
	 tx_wr_en => tx_wr_en,
    data_out => rs_encode_data,
    info => info);

-----------------------------------
-- Channel errors
-----------------------------------		

process(otuclk)
begin
if (otuclk = '1' and otuclk'event) then
      if (otu_reset = '1') then
         err_count <= (others =>'0');
      elsif (sync_start = '1') then
         err_count <= (others =>'0');
      elsif (err_count = "11111110") then
         err_count <= err_count;
      else
         err_count <= err_count + 1;
		end if;
end if;					
end process;

-----------------------------------
-- Errors select
-----------------------------------		
process(otuclk)
begin
if (otuclk = '1' and otuclk'event) then
				case error_select is
					when "00" => --no errors
						RS_ERROR_START <= "00000000";
						RS_ERROR_END <= "00000000";
						error_en <= '0';
					when "01" => --no errors
						RS_ERROR_START <= "00000000";
						RS_ERROR_END <= "00001000";-- count = 16 i.e 8 errors per channel
						error_en <= '1';
					when "10" => --no errors
						RS_ERROR_START <= "00000000";
						RS_ERROR_END <= "00001001";-- count = 18 i.e 9 errors per channel
						error_en <= '1';
					when others => 										
						RS_ERROR_START <= "00000000";
						RS_ERROR_END <= "00000000";
						error_en <= '0';
				end case;
   end if;
end process;

-----------------------------------------------------------------
--A single flop is used to store whether we want to generate an error
-----------------------------------------------------------------
process(otuclk)
begin
if (otuclk = '1' and otuclk'event) then
      if (otu_reset = '1') then
         send_error <= '0';
      elsif (sync_start = '1') then
         send_error <= '0';
      elsif ((error_en = '1') and (err_count = RS_ERROR_START)) then
         send_error <= '1';
      elsif (err_count = RS_ERROR_END) then
         send_error <= '0';
		end if;
end if;		
end process;			

----------------------------------------
-- Invert data if send_error is asserted
----------------------------------------
    rs_dec_data_in <= rs_encode_data WHEN send_error = '0' ELSE not(rs_encode_data);


-----------------------------------
-- RS Decoder and FIFO
-----------------------------------		
rs_decoder_top: rs_dec_top
	port map(
	 oduclk => oduclk,
	 otuclk => otuclk,
	 odu_reset => odu_reset,
	 otu_reset => otu_reset,
    data_in => rs_dec_data_in,
    enc_ready => enc_ready,
	 sync_start => sync_start,
    data_out => rs_data_out,
	 err_found => err_found,
    err_cnt => err_cnt,
    fail => fail,
	 ready => ready,
	 rffd => rffd,
	 rx_rd_en_valid => rx_rd_en_v);


-----------------------------------------------------
--generate the transmit address
-----------------------------------------------------
tx_ram_addr:process(odu_reset, oduclk)
begin
if odu_reset ='1' then 
     write_addr <= (others => '0');
elsif (oduclk'event and oduclk = '1') then
    if (tx_wr_en = '1') then
      write_addr <= write_addr +1;
	end if; 

end if;
end process;

-----------------------------------------------------
--generate the recieve address
-----------------------------------------------------
rx_ram_addr:process(oduclk,odu_reset)
begin
if odu_reset ='1' then 
     read_addr <= (others => '0');
elsif (oduclk'event and oduclk = '1') then
  if (rx_rd_en_v_d = '1') then
      read_addr <= read_addr +1;
end if;  
end if;
end process;



-----------------------------------------------------
--The write 1 port read 1 port
-----------------------------------------------------
process (oduclk)
begin
   if (oduclk'event and oduclk = '1') then
      if (tx_wr_en = '1') then
         delay_ram(conv_integer(write_addr)) <= FEC_sub_row_data_d;
      end if;
      read_addr_b <= read_addr;
   end if;
end process;

delay_source <= delay_ram(conv_integer(read_addr_b));


process(odu_reset, oduclk)
begin
if (oduclk'event and oduclk = '1') then
    rx_rd_en_v_d <= rx_rd_en_v;
	 rx_rd_en_v_dd <=rx_rd_en_v_d;
	 delay_source_d<=delay_source;
	 rs_data_out_d <= rs_data_out;
end if;
end process;

---------------------------------
--Process to do error checking
---------------------------------
process(oduclk,odu_reset)
begin
	if odu_reset ='1' then
		error_int <="0000";
	elsif oduclk'event and oduclk ='1' then
	if rx_rd_en_v_dd ='1' then
		if (delay_source(63 downto 0) /= rs_data_out_d(63 downto 0)) then -- does the two sets of data agree?
			error_int(0) <='1' ;
		else
			error_int(0) <='0';
		end if;	
		
		if (delay_source(127 downto 64) /= rs_data_out_d(127 downto 64)) then -- does the two sets of data agree?
			error_int(1) <='1' ;
		else
			error_int(1) <='0';
	   end if;	

		if (delay_source(191 downto 128) /= rs_data_out_d(191 downto 128)) then -- does the two sets of data agree?
			error_int(2) <='1' ;
		else
			error_int(2) <='0';
		end if;	
		
		if (delay_source(255 downto 192) /= rs_data_out_d(255 downto 192)) then -- does the two sets of data agree?
			error_int(3) <='1' ;
		else
			error_int(3) <='0';
		end if;
		end if;
	error <= error_int(0) or error_int(1) or error_int(2) or error_int(3);
	end if;
end process;



END rs_top_a;

